import asyncio
import json
import os
import time
from abc import ABC, abstractmethod
from typing import Any


class BaseConverter(ABC):
    def __init__(self):
        self.process = None
        self.start_time = None

    async def _get_file_info(self, file_path: str) -> dict:
        """使用ffprobe获取媒体文件信息"""
        try:
            probe = await asyncio.create_subprocess_exec(
                "ffprobe",
                "-v",
                "quiet",
                "-print_format",
                "json",
                "-show_format",
                "-show_streams",
                file_path,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            stdout, _ = await probe.communicate()
            return json.loads(stdout.decode())
        except Exception as e:
            print(f"Error probing file: {str(e)}")
            return {"format": {"duration": "0"}}

    async def _execute_ffmpeg_command(
        self,
        command: list,
        duration: float,
        parse_progress_func,
        progress_callback=None,
    ) -> bool:
        """执行FFmpeg命令并监控进度"""
        try:
            self.start_time = time.time()
            self.process = await asyncio.create_subprocess_exec(
                *command, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
            )

            while True:
                if not self.process or self.process.returncode is not None:
                    break

                if self.process.stdout:
                    try:
                        line = await self.process.stdout.readline()
                        if not line:
                            break

                        progress_info = parse_progress_func(line.decode(), duration)
                        if progress_info and progress_callback:
                            if isinstance(progress_info, dict):
                                progress = progress_info["progress"]
                                time_remaining = self._estimate_time_remaining(
                                    progress, duration
                                )
                                progress_callback(
                                    progress, time_remaining, progress_info
                                )
                            else:
                                time_remaining = self._estimate_time_remaining(
                                    progress_info, duration
                                )
                                progress_callback(progress_info, time_remaining)
                    except asyncio.CancelledError:
                        if self.process:
                            self.process.terminate()
                        raise

            if self.process:
                await self.process.wait()
                if self.process.returncode != 0:
                    error = (
                        await self.process.stderr.read() if self.process.stderr else b""
                    )
                    print(f"Error executing command: {error.decode()}")
                    return False

            return True

        except Exception as e:
            print(f"An error occurred: {str(e)}")
            return False
        finally:
            self.process = None

    def _check_input_file(self, input_file: str) -> bool:
        """检查输入文件是否存在"""
        if not os.path.exists(input_file):
            print(f"Error: Input file '{input_file}' does not exist")
            return False
        return True

    def _ensure_output_format(self, output_file: str, output_format: str) -> str:
        """确保输出文件具有正确的扩展名"""
        if not output_file.endswith(f".{output_format}"):
            output_file = f"{output_file}.{output_format}"
        return output_file

    @abstractmethod
    def _parse_progress(self, line: str, duration: float) -> Any:
        """解析FFmpeg进度输出的抽象方法"""
        pass

    @abstractmethod
    def _estimate_time_remaining(self, progress: float, duration: float) -> str:
        """根据当前进度估算剩余时间的抽象方法"""
        pass

    @abstractmethod
    async def convert(
        self, input_file: str, output_file: str, output_format: str, **kwargs
    ) -> bool:
        """转换媒体文件的抽象方法"""
        pass



import time
from collections.abc import Callable
from datetime import timedelta
from typing import Any

from src.base import BaseConverter


class AudioConverter(BaseConverter):
    def _parse_progress(self, line: str, duration: float) -> dict[str, Any] | None:
        """解析FFmpeg进度输出"""
        if "time=" in line:
            try:
                time_str = line.split("time=")[1].split()[0]
                if ":" in time_str:
                    h, m, s = time_str.split(":")
                    time_seconds = float(h) * 3600 + float(m) * 60 + float(s)
                else:
                    time_seconds = float(time_str)
                progress = (time_seconds / duration) * 100
                return {
                    "progress": progress,
                    "time": time_seconds,
                    "duration": duration,
                }
            except (ValueError, IndexError):
                return None
        return None

    def _estimate_time_remaining(self, progress: float, duration: float) -> str:
        """根据当前进度估算剩余时间"""
        if progress <= 0:
            return "计算中..."

        elapsed_time = time.time() - (
            self.start_time if self.start_time is not None else time.time()
        )
        total_time = (elapsed_time * 100) / progress
        remaining_seconds = total_time - elapsed_time

        remaining_time = timedelta(seconds=int(remaining_seconds))
        return str(remaining_time)

    async def convert(
        self,
        input_file: str,
        output_file: str,
        output_format: str,
        sample_rate: int | None = None,
        channels: int | None = None,
        progress_callback: Callable[[float, str, dict[str, Any]], None] | None = None,
        **kwargs,
    ) -> bool:
        """使用FFmpeg将音频文件转换为指定格式，并提供进度监控

        Args:
            input_file (str): 输入音频文件路径
            output_file (str): 输出音频文件路径
            output_format (str): 目标输出格式（如'mp3', 'wav', 'ogg'等）
            sample_rate (int, optional): 采样率（Hz，如44100）
            channels (int, optional): 音频通道数（如2表示立体声）
            progress_callback (callable, optional): 进度回调函数，接收进度百分比和
            剩余时间
            **kwargs: 其他可选参数

        Returns:
            bool: 转换成功返回True，否则返回False
        """
        if not self._check_input_file(input_file):
            return False

        probe = await self._get_file_info(input_file)
        duration = float(probe["format"]["duration"])

        command = ["ffmpeg", "-i", input_file]

        if sample_rate:
            command.extend(["-ar", str(sample_rate)])

        if channels:
            command.extend(["-ac", str(channels)])

        command.extend(["-progress", "pipe:1"])
        output_file = self._ensure_output_format(output_file, output_format)
        command.extend(["-y", output_file])

        success = await self._execute_ffmpeg_command(
            command, duration, self._parse_progress, progress_callback
        )

        if success and progress_callback:
            progress_callback(100, "完成", {"status": "finished"})
            print(f"\nSuccessfully converted {input_file} to {output_file}")

        return success



import os

import pytest

from src.video import VideoConverter

# 测试文件路径
TEST_DIR = os.path.dirname(os.path.abspath(__file__))
SAMPLE_VIDEO = os.path.join(TEST_DIR, "samples", "test_video.mov")

# 创建测试所需的目录
os.makedirs(os.path.join(TEST_DIR, "samples"), exist_ok=True)
os.makedirs(os.path.join(TEST_DIR, "output"), exist_ok=True)


@pytest.fixture(scope="session")
def video_converter() -> VideoConverter:
    return VideoConverter()


@pytest.mark.asyncio
@pytest.mark.parametrize("fmt", ["mp4", "mkv"])
async def test_video_format_conversion(video_converter: VideoConverter, fmt):
    """测试主要视频格式转换"""
    output_file = os.path.join(TEST_DIR, "output", f"test_output.{fmt}")
    result = await video_converter.convert(
        input_file=SAMPLE_VIDEO, output_file=output_file, output_format=fmt
    )
    assert result is True
    assert os.path.exists(output_file)


@pytest.mark.asyncio
@pytest.mark.parametrize("res", ["1280x720", "1920x1080"])
async def test_video_resolution_conversion(video_converter: VideoConverter, res):
    """测试主要分辨率转换"""
    output_file = os.path.join(TEST_DIR, "output", f"test_output_{res}.mp4")
    result = await video_converter.convert(
        input_file=SAMPLE_VIDEO,
        output_file=output_file,
        output_format="mp4",
        resolution=res,
    )
    assert result is True
    assert os.path.exists(output_file)


@pytest.mark.asyncio
async def test_video_basic_conversion(video_converter: VideoConverter):
    """测试基本视频转换功能"""
    output_file = os.path.join(TEST_DIR, "output", "test_output.mkv")

    progress_data = []

    def progress_callback(progress, time_remaining, info):
        progress_data.append(
            {"progress": progress, "time_remaining": time_remaining, "info": info}
        )

    result = await video_converter.convert(
        input_file=SAMPLE_VIDEO,
        output_file=output_file,
        output_format="mkv",
        video_codec="h264",
        video_bitrate="5M",
        resolution="1280x720",
        fps=30,
        audio_codec="aac",
        audio_bitrate="192k",
        preset="fast",
        progress_callback=progress_callback,
    )

    assert result is True
    assert os.path.exists(output_file)
    assert len(progress_data) > 0
    assert progress_data[-1]["progress"] == 100.0


@pytest.mark.asyncio
async def test_video_conversion_errors(video_converter: VideoConverter):
    """测试视频转换错误处理"""
    # 测试不存在的输入文件
    result = await video_converter.convert(
        input_file="nonexistent.mp4", output_file="output.mkv", output_format="mkv"
    )
    assert result is False


@pytest.mark.asyncio
async def test_video_audio_params(video_converter: VideoConverter):
    """测试视频音频参数设置"""
    output_file = os.path.join(TEST_DIR, "output", "test_output_audio.mp4")
    result = await video_converter.convert(
        input_file=SAMPLE_VIDEO,
        output_file=output_file,
        output_format="mp4",
        audio_codec="aac",
        audio_bitrate="192k",
    )
    assert result is True
    assert os.path.exists(output_file)


# @pytest.mark.asyncio
# async def test_video_conversion_interrupt(video_converter: VideoConverter):
#     """测试视频转换中断处理"""
#     output_file = os.path.join(TEST_DIR, "output", "test_output_interrupt.mp4")

#     def progress_callback(progress, time_remaining, info):
#         if progress > 50:  # 当进度超过50%时中断
#             if video_converter.process and video_converter.process.returncode is None:
#                 video_converter.process.terminate()

#     result = await video_converter.convert(
#         input_file=SAMPLE_VIDEO,
#         output_file=output_file,
#         output_format="mp4",
#         progress_callback=progress_callback,
#     )
#     assert result is False



import time
from collections.abc import Callable
from datetime import timedelta
from typing import Any

from src.base import BaseConverter


class VideoConverter(BaseConverter):
    def _parse_progress(self, line: str, duration: float) -> dict[str, Any] | None:
        """解析FFmpeg进度输出"""
        progress_info = {}

        try:
            for item in line.strip().split():
                if "=" in item:
                    key, value = item.split("=", 1)
                    progress_info[key] = value

            if "out_time_ms" in progress_info:
                time_ms = int(progress_info["out_time_ms"])
                progress = (time_ms / 1000000 / duration) * 100
                progress_info["progress"] = progress

                if "frame" in progress_info:
                    progress_info["current_frame"] = int(progress_info["frame"])

                if "speed" in progress_info:
                    progress_info["conversion_speed"] = progress_info["speed"]

                return progress_info

        except Exception:
            pass

        return None

    def _estimate_time_remaining(self, progress: float, duration: float) -> str:
        """根据当前进度估算剩余时间"""
        if progress <= 0:
            return "计算中..."

        elapsed_time = time.time() - (self.start_time or time.time())
        total_time = (elapsed_time * 100) / progress
        remaining_seconds = total_time - elapsed_time

        remaining_time = timedelta(seconds=int(remaining_seconds))
        return str(remaining_time)

    async def convert(
        self,
        input_file: str,
        output_file: str,
        output_format: str,
        video_codec: str | None = None,
        video_bitrate: str | None = None,
        resolution: str | None = None,
        fps: int | None = None,
        audio_codec: str | None = None,
        audio_bitrate: str | None = None,
        preset: str = "medium",
        crf: int | None = None,
        progress_callback: Callable[[float, str, dict[str, Any]], None] | None = None,
        **kwargs,
    ) -> bool:
        """使用FFmpeg转换视频文件到指定格式，并提供进度监控

        Args:
            input_file (str): 输入视频文件路径
            output_file (str): 输出视频文件路径
            output_format (str): 目标输出格式（如'mp4', 'mkv'等）
            video_codec (str, optional): 视频编码器（如'h264', 'h265'等）
            video_bitrate (str, optional): 视频比特率（如'5M'）
            resolution (str, optional): 视频分辨率（如'1920x1080'）
            fps (int, optional): 视频帧率
            audio_codec (str, optional): 音频编码器（如'aac'）
            audio_bitrate (str, optional): 音频比特率（如'192k'）
            preset (str, optional): 编码器预设值（如'fast', 'medium'等）。默认为'medium'
            crf (int, optional): 恒定速率因子，控制视频质量（0-51，值越小质量越好）
            progress_callback (callable, optional): 进度回调函数，接收进度百分比、
                剩余时间和详细信息
            **kwargs: 其他可选参数

        Returns:
            bool: 转换成功返回True，否则返回False
        """
        if not self._check_input_file(input_file):
            return False

        probe = await self._get_file_info(input_file)
        duration = float(probe["format"]["duration"])

        command = ["ffmpeg", "-i", input_file]

        if video_codec:
            command.extend(["-c:v", video_codec])

        if video_bitrate:
            command.extend(["-b:v", video_bitrate])

        if resolution:
            command.extend(["-s", resolution])

        if fps:
            command.extend(["-r", str(fps)])

        if audio_codec:
            command.extend(["-c:a", audio_codec])

        if audio_bitrate:
            command.extend(["-b:a", audio_bitrate])

        if video_codec in ["h264", "h265", "hevc"]:
            command.extend(["-preset", preset])
            if crf is not None:
                command.extend(["-crf", str(crf)])

        command.extend(["-progress", "pipe:1"])
        output_file = self._ensure_output_format(output_file, output_format)
        command.extend(["-y", output_file])

        success = await self._execute_ffmpeg_command(
            command, duration, self._parse_progress, progress_callback
        )

        if success and progress_callback:
            progress_callback(100, "完成", {"status": "finished"})
            print(f"\nSuccessfully converted {input_file} to {output_file}")

        return success



import os

import pytest

from src.audio import AudioConverter

# 测试文件路径
TEST_DIR = os.path.dirname(os.path.abspath(__file__))
SAMPLE_AUDIO = os.path.join(TEST_DIR, "samples", "test_audio.mp3")

# 创建测试所需的目录
os.makedirs(os.path.join(TEST_DIR, "samples"), exist_ok=True)
os.makedirs(os.path.join(TEST_DIR, "output"), exist_ok=True)


@pytest.fixture(scope="session")
def audio_converter() -> AudioConverter:
    return AudioConverter()


@pytest.mark.asyncio
@pytest.mark.parametrize("fmt", ["mp3", "wav"])
async def test_audio_format_conversion(audio_converter: AudioConverter, fmt):
    """测试主要音频格式转换"""
    output_file = os.path.join(TEST_DIR, "output", f"test_output.{fmt}")
    result = await audio_converter.convert(
        input_file=SAMPLE_AUDIO, output_file=output_file, output_format=fmt
    )
    assert result is True
    assert os.path.exists(output_file)


@pytest.mark.asyncio
@pytest.mark.parametrize("rate", [44100, 48000])
async def test_audio_sample_rate_conversion(audio_converter: AudioConverter, rate):
    """测试关键采样率转换"""
    output_file = os.path.join(TEST_DIR, "output", f"test_output_{rate}.wav")
    result = await audio_converter.convert(
        input_file=SAMPLE_AUDIO,
        output_file=output_file,
        output_format="wav",
        sample_rate=rate,
    )
    assert result is True
    assert os.path.exists(output_file)


@pytest.mark.asyncio
async def test_audio_basic_conversion(audio_converter: AudioConverter):
    """测试基本音频转换功能"""
    output_file = os.path.join(TEST_DIR, "output", "test_output.wav")

    progress_data = []

    def progress_callback(progress, time_remaining, info):
        progress_data.append(
            {"progress": progress, "time_remaining": time_remaining, "info": info}
        )

    result = await audio_converter.convert(
        input_file=SAMPLE_AUDIO,
        output_file=output_file,
        output_format="wav",
        sample_rate=44100,
        channels=2,
        progress_callback=progress_callback,
    )

    assert result is True
    assert os.path.exists(output_file)
    assert len(progress_data) > 0
    assert progress_data[-1]["progress"] == 100.0


@pytest.mark.asyncio
async def test_audio_conversion_errors(audio_converter: AudioConverter):
    """测试音频转换错误处理"""
    # 测试不存在的输入文件
    result = await audio_converter.convert(
        input_file="nonexistent.mp3", output_file="output.wav", output_format="wav"
    )
    assert result is False


@pytest.mark.asyncio
@pytest.mark.parametrize("channels", [1, 2])
async def test_audio_channels_conversion(audio_converter: AudioConverter, channels):
    """测试音频声道数转换"""
    output_file = os.path.join(TEST_DIR, "output", f"test_output_{channels}ch.wav")
    result = await audio_converter.convert(
        input_file=SAMPLE_AUDIO,
        output_file=output_file,
        output_format="wav",
        channels=channels,
    )
    assert result is True
    assert os.path.exists(output_file)


# @pytest.mark.asyncio
# async def test_audio_conversion_interrupt(audio_converter: AudioConverter):
#     """测试音频转换中断处理"""
#     output_file = os.path.join(TEST_DIR, "output", "test_output_interrupt.wav")

#     def progress_callback(progress, time_remaining, info):
#         if progress > 5:  # 当进度超过5%时中断
#             if audio_converter.process and audio_converter.process.returncode is None:
#                 audio_converter.process.terminate()

#     result = await audio_converter.convert(
#         input_file=SAMPLE_AUDIO,
#         output_file=output_file,
#         output_format="wav",
#         progress_callback=progress_callback,
#         sample_rate=44100,
#         channels=2,
#     )
#     assert result is False
